#!/bin/bash
#<pre><code>

# What does this do.. and why?
# (and why is it so complicated?)
#
# What this does
#
#   1. Sets some vars and asks some questions
#   2. Installs everything into a single place
#      (inculding deps like node.js, with the correct version)
#   3. Depending on OS, creates a user for the service
#   4. Depending on OS, register with system launcher
#
# Why
#
#   So that you can get a fully configured, running product,
#   with zero manual configuration in a matter of seconds -
#   and have an uninstall that's just as easy.
#
# Why so complicated?
#
#  To support nuance differences between various versions of
#  Linux, macOS, and Android, including whether it's being
#  installed with user privileges, as root, wit a system user
#  system daemon launcher, etc. Also, this is designed to be
#  reusable with many apps and services, so it's very variabled...

set -e
set -u

### http_bash exported by get.sh

TELEBIT_VERSION=${TELEBIT_VERSION:-master}
my_email=${1:-}
my_relay=${2:-}
my_servernames=${3:-}
my_secret=${4:-}
my_user="telebit"
my_app_pkg_name="cloud.telebit.remote"
my_app="telebit"
my_daemon="telebitd"
my_bin="telebit.js"
my_name="Telebit Remote"
my_repo="telebit.js"
my_root=${my_root:-} # todo better install script
sudo_cmd="sudo"
sudo_cmde="sudo "
exec 3<>/dev/tty
read_cmd="read -u 3"
# TODO detect if rsync is available and use rsync -a (more portable)
rsync_cmd="cp -pPR"

set +e
my_edit=$(basename "${EDITOR:-}")
if [ -z "$my_edit" ]; then
  my_edit=$(basename "$(type -p edit)")
fi
if [ -z "$my_edit" ]; then
  my_edit=$(basename "$(type -p nano)")
fi
if [ -z "$my_edit" ]; then
  my_edit=$(basename "$(type -p vim)")
fi
if [ -z "$my_edit" ]; then
  my_edit=$(basename "$(type -p vi)")
fi
if [ -z "$my_edit" ]; then
  my_edit="nano"
fi
set -e

if [ "root" == $(whoami) ] || [ 0 == $(id -u) ]; then
  sudo_cmd=" "
  sudo_cmde=""
fi

echo ""

if [ -z "${TELEBIT_PATH:-}" ]; then
  echo 'TELEBIT_PATH="'${TELEBIT_PATH:-}'"'
  TELEBIT_PATH=/opt/$my_app
fi

echo "Installing $my_name to '$TELEBIT_PATH'"
# v10.2+ has much needed networking fixes, but breaks ursa. v9.x has severe networking bugs. v8.x has working ursa, but requires tls workarounds"
NODEJS_VER="${NODEJS_VER:-v10.2}"
export NODEJS_VER
export NODE_PATH="$TELEBIT_PATH/lib/node_modules"
export NPM_CONFIG_PREFIX="$TELEBIT_PATH"
export PATH="$TELEBIT_PATH/bin:$PATH"
sleep 0.25
echo "(your password may be required to complete installation)"

#echo "${sudo_cmde}mkdir -p '$TELEBIT_PATH'{etc,var/log}"
$sudo_cmd mkdir -p "$TELEBIT_PATH"

echo "  - installing node.js runtime to '$TELEBIT_PATH'..."
http_bash https://git.coolaj86.com/coolaj86/node-installer.sh/raw/branch/master/install.sh --no-dev-deps >/dev/null 2>/dev/null

my_node="$TELEBIT_PATH/bin/node"
my_npm="$my_node $TELEBIT_PATH/bin/npm"
my_tmp="$(mktemp -d)"
mkdir -p $my_tmp

$sudo_cmd mkdir -p "$TELEBIT_PATH/etc"
$sudo_cmd mkdir -p "$TELEBIT_PATH/var/log"
$sudo_cmd chown -R $(id -u -n):$(id -g -n) "$TELEBIT_PATH"
#echo "${sudo_cmde}mkdir -p '/etc/$my_app/'"
#$sudo_cmd mkdir -p "/etc/$my_app/"
#$sudo_cmd chown $(id -u -n):$(id -g -n) "/etc/$my_app/"

#https://git.coolaj86.com/coolaj86/telebit.js.git
#https://git.coolaj86.com/coolaj86/telebit.js/archive/:tree:.tar.gz
#https://git.coolaj86.com/coolaj86/telebit.js/archive/:tree:.zip
set +e
my_unzip=$(type -p unzip)
my_tar=$(type -p tar)
# TODO extract to temporary directory, configure, copy etc, replace
if [ -n "$my_unzip" ]; then
  rm -f $my_tmp/$my_app-$TELEBIT_VERSION.zip
  echo "  - installing telebit zip to '$TELEBIT_PATH'..."
  http_get https://git.coolaj86.com/coolaj86/$my_repo/archive/$TELEBIT_VERSION.zip $my_tmp/$my_app-$TELEBIT_VERSION.zip
  # -o means overwrite, and there is no option to strip
  $my_unzip -o $my_tmp/$my_app-$TELEBIT_VERSION.zip -d $TELEBIT_PATH/ >/dev/null
  $rsync_cmd  $TELEBIT_PATH/$my_repo/* $TELEBIT_PATH/ > /dev/null
  rm -rf $TELEBIT_PATH/$my_repo
elif [ -n "$my_tar" ]; then
  rm -f $my_tmp/$my_app-$TELEBIT_VERSION.tar.gz
  echo "  - installing telebit tar.gz to '$TELEBIT_PATH'..."
  http_get https://git.coolaj86.com/coolaj86/$my_repo/archive/$TELEBIT_VERSION.tar.gz $my_tmp/$my_app-$TELEBIT_VERSION.tar.gz
  $my_tar -xzf $my_tmp/$my_app-$TELEBIT_VERSION.tar.gz --strip 1 -C $TELEBIT_PATH/ >/dev/null
else
  echo "Neither tar nor unzip found. Abort."
  exit 13
fi
set -e

pushd $TELEBIT_PATH >/dev/null
  echo "  - installing telebit npm dependencies to '$TELEBIT_PATH'..."
  echo "    (are you noticing a pattern of where things are installed?)"
  $my_npm install >/dev/null 2>/dev/null
popd >/dev/null

echo "  - configuring telebit..."
echo ""

# telebit remote
echo '#!/bin/bash' > "$TELEBIT_PATH/bin/$my_app"
echo "$my_node $TELEBIT_PATH/bin/$my_bin "'"$@"' >> "$TELEBIT_PATH/bin/$my_app"
chmod a+x "$TELEBIT_PATH/bin/$my_app"

# telebit daemon
echo '#!/bin/bash' > "$TELEBIT_PATH/bin/$my_daemon"
echo "$my_node $TELEBIT_PATH/bin/$my_daemon.js daemon "'"$@"' >> "$TELEBIT_PATH/bin/$my_daemon"
chmod a+x "$TELEBIT_PATH/bin/$my_daemon"

# Create uninstall script based on the install script variables
cat << EOF > $TELEBIT_PATH/bin/${my_app}_uninstall
#!/bin/bash
if [ "$(type -p launchctl)" ]; then
  sudo launchctl unload -w /Library/LaunchDaemons/${my_app_pkg_name}.plist
  sudo rm -rf /Library/LaunchDaemons/cloud.telebit.remote.plist
fi
if [ "$(type -p systemctl)" ]; then
  sudo systemctl disable telebit >/dev/null; sudo systemctl stop telebit
  sudo rm -rf /etc/systemd/system/$my_app.service
fi
sudo rm -rf $TELEBIT_PATH /usr/local/bin/$my_app
sudo rm -rf $TELEBIT_PATH /usr/local/bin/$my_daemon
rm -rf ~/.config/$my_app ~/.local/share/$my_app
EOF
chmod a+x $TELEBIT_PATH/bin/${my_app}_uninstall

echo "    > ${sudo_cmde}ln -sf $TELEBIT_PATH/bin/$my_app /usr/local/bin/$my_app"
$sudo_cmd ln -sf $TELEBIT_PATH/bin/$my_app /usr/local/bin/$my_app
echo "    > ${sudo_cmde}ln -sf $TELEBIT_PATH/bin/$my_daemon /usr/local/bin/$my_daemon"
$sudo_cmd ln -sf $TELEBIT_PATH/bin/$my_daemon /usr/local/bin/$my_daemon

set +e
if type -p setcap >/dev/null 2>&1; then
  #echo "Setting permissions to allow $my_app to run on port 80 and port 443 without sudo or root"
  echo "    > ${sudo_cmde}setcap cap_net_bind_service=+ep $TELEBIT_PATH/bin/node"
  $sudo_cmd setcap cap_net_bind_service=+ep $TELEBIT_PATH/bin/node
fi
set -e

set +e
# TODO for macOS https://apple.stackexchange.com/questions/286749/how-to-add-a-user-from-the-command-line-in-macos
if type -p adduser >/dev/null 2>/dev/null; then
  if [ -z "$(cat $my_root/etc/passwd | grep $my_user)" ]; then
    $sudo_cmd adduser --home $TELEBIT_PATH --gecos '' --disabled-password $my_user >/dev/null 2>&1
  fi
  #my_user=$my_app_name
  my_group=$my_user
elif [ -n "$(cat /etc/passwd | grep www-data:)" ]; then
  # Linux (Ubuntu)
  my_user=www-data
  my_group=www-data
elif [ -n "$(cat /etc/passwd | grep _www:)" ]; then
  # Mac
  my_user=_www
  my_group=_www
else
  # Unsure
  my_user=$(id -u -n) # $(whoami)
  my_group=$(id -g -n)
fi
set -e

# ~/.config/systemd/user/
# %h/.config/telebit/telebit.yml
echo "  - adding $my_app as a system service"
# TODO detect with type -p
my_system_launcher=""
if [ -d "/Library/LaunchDaemons" ]; then
  my_system_launcher="launchd"
  my_app_launchd_service="Library/LaunchDaemons/${my_app_pkg_name}.plist"
  echo "    > ${sudo_cmde}$rsync_cmd $TELEBIT_PATH/usr/share/dist/$my_app_launchd_service /$my_app_launchd_service"
  $sudo_cmd $rsync_cmd "$TELEBIT_PATH/usr/share/dist/$my_app_launchd_service" "/$my_app_launchd_service"

  echo "    > ${sudo_cmde}chown root:wheel $my_root/$my_app_launchd_service"
  $sudo_cmd chown root:wheel "$my_root/$my_app_launchd_service"
  echo "    > ${sudo_cmde}launchctl unload -w $my_root/$my_app_launchd_service >/dev/null 2>/dev/null"
  $sudo_cmd launchctl unload -w "$my_root/$my_app_launchd_service" >/dev/null 2>/dev/null

elif [ -d "$my_root/etc/systemd/system" ]; then
  my_system_launcher="systemd"
  echo "    > ${sudo_cmde}$rsync_cmd $TELEBIT_PATH/usr/share/dist/etc/systemd/system/$my_app.service /etc/systemd/system/$my_app.service"
  $sudo_cmd $rsync_cmd "$TELEBIT_PATH/usr/share/dist/etc/systemd/system/$my_app.service" "/etc/systemd/system/$my_app.service"

  $sudo_cmd systemctl daemon-reload
  echo "    > ${sudo_cmde}systemctl enable $my_app"
  $sudo_cmd systemctl enable $my_app >/dev/null
fi

sleep 1
echo ""
echo ""
echo "=============================================="
echo "            Launcher Configuration            "
echo "=============================================="
echo ""

my_stopper=""
if [ "systemd" == "$my_system_launcher" ]; then

  my_stopper="${sudo_cmde}systemctl stop $my_app"
  echo "Edit the config and restart, if desired:"
  echo ""
  echo "    ${sudo_cmde}$my_edit $TELEBIT_PATH/etc/$my_app.yml"
  echo "    ${sudo_cmde}systemctl restart $my_app"
  echo ""
  echo "Or disabled the service and start manually:"
  echo ""
  echo "    ${sudo_cmde}systemctl stop $my_app"
  echo "    ${sudo_cmde}systemctl disable $my_app"
  echo "    $my_daemon --config $TELEBIT_PATH/etc/$my_daemon.yml"

elif [ "launchd" == "$my_system_launcher" ]; then

  my_stopper="${sudo_cmde}launchctl unload $my_root/$my_app_launchd_service"
  echo "Edit the config and restart, if desired:"
  echo ""
  echo "    ${sudo_cmde}$my_edit $TELEBIT_PATH/etc/$my_app.yml"
  echo "    ${sudo_cmde}launchctl unload $my_root/$my_app_launchd_service"
  echo "    ${sudo_cmde}launchctl load -w $my_root/$my_app_launchd_service"
  echo ""
  echo "Or disabled the service and start manually:"
  echo ""
  echo "    ${sudo_cmde}launchctl unload -w $my_root/$my_app_launchd_service"
  echo "    $my_daemon --config $TELEBIT_PATH/etc/$my_daemon.yml"

else

  my_stopper="not started"
  echo "Edit the config, if desired:"
  echo ""
  echo "    ${sudo_cmde}$my_edit $my_config"
  echo ""
  echo "Run the service manually (we couldn't detect your system service to do that automatically):"
  echo ""
  echo "    $my_daemon --config $my_config"

fi

sleep 2

# TODO don't create this in TMP_PATH if it exists in TELEBIT_PATH
my_config="$TELEBIT_PATH/etc/$my_daemon.yml"
mkdir -p "$(dirname $my_config)"
if [ ! -e "$my_config" ]; then

  echo "sock: $TELEBIT_PATH/var/run/telebit.sock" >> "$my_config"
  cat $TELEBIT_PATH/usr/share/$my_daemon.tpl.yml >> "$my_config"

fi

my_config="$HOME/.config/$my_app/$my_app.yml"
mkdir -p "$(dirname $my_config)"
if [ ! -e "$my_config" ]; then

  echo "sock: $TELEBIT_PATH/var/run/telebit.sock" >> "$my_config"

fi

#echo "${sudo_cmde}chown -R $my_user '$TELEBIT_PATH'
$sudo_cmd chown -R $my_user "$TELEBIT_PATH"

###############################
# Actually Launch the Service #
###############################
echo ""
if [ "launchd" == "$my_system_launcher" ]; then
  echo "  > ${sudo_cmde}launchctl load -w $my_root/$my_app_launchd_service"
  $sudo_cmd launchctl load -w "$my_root/$my_app_launchd_service"
fi
if [ "systemd" == "$my_system_launcher" ]; then
  echo "  > ${sudo_cmde}systemctl start $my_app"
  $sudo_cmd systemctl restart $my_app
fi

echo "  > telebit init --tty"
echo ""
sleep 0.25

$TELEBIT_PATH/bin/node $TELEBIT_PATH/bin/telebit.js init --tty

$TELEBIT_PATH/bin/node $TELEBIT_PATH/bin/telebit.js enable

